""" Import libraries """
import cv2
import os
import numpy as np
import matplotlib.pyplot as plt
import math
import csv
import time
import ast
import seaborn as sns
from statistics import mean
from collections import Counter

"""
This script consists of multiple functions created in order to track objects over time. Image processing techniques are 
used before feature detection takes place. The center coordinates of the detected objects are used together with the 
Eucleadian distance to track the objects. The code works best if there are few lice in the images. 

To use this code, the name of the input files need to be on the form: "Tracking-Color-Intensity-Pulsation-Replicate", 
where color can be Green, Blue or White, intensity can be O1, OD3 or OD5, pulsation can be P0, P1, P2, P3, and 
Replicate can be R1, R2 or R3. Tracking is used to distinguish the experiments from the detection experiments, 
where a T is placed in front of the name. One example is T_Green_OD1_P1_R1. 

In main() at the bottom of this file, add all the videos you want to analyze to "list_videos". "path" and "path_res"
also need to be updated. In addition, the length [mm] of each pixel in x- and y-direction needs to be specified. 

The outputs of the code are .csv-files containing all the center coordinates of the detection ("centers.csv"), the 
associated tracks ("tracks.csv"), corresponding data ("data.csv"), as well as velocity in both x- and y-direction and 
total velocity ("Speeds_TOT.csv", "Speeds_X.csv", "Speeds_Y.csv"). In addition, plots showing the tracks and speeds of 
every track is outputted. 

For more details about the method, see the thesis. 
"""


def get_information(video):
    """
    Retrieve general information.
    :return: number of frames in video
    """
    video = video
    path = "W:\TrackingVideos" + "/" + video
    frames = os.listdir(path)
    frames.sort()
    num_frames = len(frames)

    print("get information")
    return num_frames


def calculate_median(video):
    """
    Calculate median pixel values in the video to use for median subtraction.
    :return: median, median bright frames, median dark frames, maximum mean pixel value in video, minimum pixel value in video
    """

    start = time.time()

    video = video

    path = "W:\TrackingVideos/" + video

    frames = os.listdir(path)

    frames.sort()

    num_frames = len(frames)

    """ Calculate median of all frames """
    median_frames = []
    mean_frames = []

    for i in range(num_frames - 1000, num_frames):
        if os.path.isdir(os.path.join(path, frames[i])):
            continue

        med_path = os.path.join(path, frames[i])

        with open(med_path, "rb") as frame:
            med_raw = frame.read()
            med_frame = np.frombuffer(med_raw, dtype=np.uint8).reshape(1560, 2200)
            med_frame = med_frame[200:1400, 315:2095]
            median_frames.append(med_frame)
            mean_frames.append(med_frame.mean())

    median = np.median(median_frames, axis=0).astype(np.uint8)

    max_mean = np.max(mean_frames)
    min_mean = np.min(mean_frames)

    diff = max_mean - min_mean

    a = video.split("_")
    pulse_type = a[3]

    if pulse_type == "P0":
        cons = diff * 0.4
    elif pulse_type == "P1":
        cons = diff * 0.15
    elif pulse_type == "P2":
        cons = diff * 0.55
    elif pulse_type == "P3":
        cons = diff * 0.4

    """ Calculate separate median for bright and dark frames """
    median_frames_bright = []
    median_frames_dark = []
    for i in range(num_frames - 1000, num_frames):
        if os.path.isdir(os.path.join(path, frames[i])):
            continue
        med_path = os.path.join(path, frames[i])
        with open(med_path, "rb") as file:
            med_raw = file.read()
            med_frame = np.frombuffer(med_raw, dtype=np.uint8).reshape(1560, 2200)
            med_frame = med_frame[200:1400, 315:2095]

            mean = med_frame.mean()
            med_frame = med_frame

            if mean >= min_mean + cons:
                median_frames_bright.append(med_frame)
            else:
                median_frames_dark.append(med_frame)

    median_bright = np.median(median_frames_bright, axis=0)
    median_dark = np.median(median_frames_dark, axis=0)

    median_bright = median_bright.astype(np.uint8)
    median_dark = median_dark.astype(np.uint8)

    end = time.time()

    print("Time median calculation: ", end - start)

    return median, median_bright, median_dark, min_mean, max_mean


def calculate_median_p0(video):
    """
    Calculates median pixel values for pulsation 0 (no pulsation)
    :return: median for both dark and bright frames 
    """
    video = video

    path = "W:\TrackingVideos/" + video

    frames = os.listdir(path)

    frames.sort()

    median_frames_dark = []
    median_frames_light = []

    for i in range(1, 1000):
        if os.path.isdir(path + "/" + frames[i]):
            continue

        med_path_dark = os.path.join(path, frames[i])
        med_path_light = os.path.join(path, frames[-i])

        with open(med_path_dark, "rb") as frame_dark:
            med_raw_dark = frame_dark.read()
            med_frame_dark = np.frombuffer(med_raw_dark, dtype=np.uint8).reshape(1560, 2200)
            med_frame_dark = med_frame_dark[200:1400, 315:2095]
            median_frames_dark.append(med_frame_dark)

        with open(med_path_light, "rb") as frame_light:
            med_raw_light = frame_light.read()
            med_frame_light = np.frombuffer(med_raw_light, dtype=np.uint8).reshape(1560, 2200)
            med_frame_light = med_frame_light[200:1400, 315:2095]
            median_frames_light.append(med_frame_light)

    median_dark = np.median(median_frames_dark, axis=0).astype(np.uint8)
    median_light = np.median(median_frames_light, axis=0).astype(np.uint8)

    return median_dark, median_light


def choose_median_p0(i, median_dark, median_light):
    """
    Choose the appropriate median for the current frame. 
    :param i: frame number
    :param median_dark: median for dark frames
    :param median_light: median for bright frames
    :return: median 
    """
    if i < 128 * 30:
        median = median_dark
    else:
        median = median_light
    return median


def choose_median(frame, median_bright, median_dark, min_mean, max_mean, video):
    """
    Choose which median to use for median subtraction.
    :param frame: current video frame
    :param median_bright: median calculated for frames when light source is turned on.
    :param median_dark: median calculated for frames when light source is turned off.
    :param min_mean: smallest mean value of all pixels in a video frame.
    :param max_mean: largest mean value of all pixels in a video frame.
    :param video: video which is being analyzed
    :return: the median to use for median subtraction.
    """

    mean = frame.mean()

    diff = max_mean - min_mean

    a = video.split("_")
    pulse_type = a[3]

    if pulse_type == "P0":
        cons = diff * 0.4
    elif pulse_type == "P1":
        cons = diff * 0.15
    elif pulse_type == "P2":
        cons = diff * 0.55
    elif pulse_type == "P3":
        cons = diff * 0.4

    if mean > min_mean + cons:
        median = median_bright

    else:
        median = median_dark

    return median


def median_subtraction(frame, median):
    """
    Subtracts pixel median from frame to remove background.
    :param frame: current video frame.
    :param median: calculated pixel medians of video frames.
    :return: frame with the median subtracted.
    """
    im = frame.astype(np.int16)
    im -= median
    im = im.clip(0, 255).astype(np.uint8)

    return im


def calculate_thresh(frame, x, n):
    """
    Calculates the threshold used for thresholding the frame in order for the objects to stand out from the background.
    :param frame: current video frame.
    :param x: initial threshold value.
    :param n: number of objects in the tank to be detected.
    :return: updated threshold value.
    """
    # Maria har 25 som initverdi
    n_detections = 0
    count = 0
    while not 2 < n_detections <= n and count < 240:
        ret, thresh = cv2.threshold(frame, x, 255, 0)
        kernel2 = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))
        dilation = cv2.dilate(thresh, kernel2, iterations=3)

        contours, hierarchy = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        n_det = 0
        for (_, contour) in enumerate(contours):
            m = cv2.moments(contour)
            area = cv2.contourArea(contour)
            if 60 < area < 300:
                n_det += 1
        n_detections = n_det
        x += 1
        count += 1

    return x


def detect(frame, median):
    """Detect objects in video frame using following pipeline
        - Convert captured frame from BGR to GRAY
        - Perform Background Subtraction
        - Detect edges using Canny Edge Detection
          http://docs.opencv.org/trunk/da/d22/tutorial_py_canny.html
        - Retain only edges within the threshold
        - Find contours
        - Find centroids for each valid contours
    Args:
        frame: single video frame
        median: median used for median subtraction 
    Return:
        centers: vector of object centroids in a frame
    """
    im = frame

    median_sub = median_subtraction(im, median)

    # Gauss
    im_gauss = cv2.GaussianBlur(median_sub, (3, 3), 0)

    # Retain only edges within the threshold
    threshold = calculate_thresh(im_gauss, 10, 15)

    ret, thresh = cv2.threshold(im_gauss, threshold, 255, 0)

    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (3, 3))

    dilation = cv2.dilate(thresh, kernel, iterations=3)

    # Find contours
    contours, hierarchy = cv2.findContours(dilation, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    color_img = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)

    centers = []

    used_contours = []

    for cnt in contours:
        try:
            # Calculate and draw circle
            (x, y), radius = cv2.minEnclosingCircle(cnt)
            centeroid = (int(x), int(y))
            radius = int(radius)
            area = cv2.contourArea(cnt)
            if 60 < area < 300:
                cv2.circle(color_img, centeroid, radius + 10, (0, 0, 255), 2)
                centers.append([x, y])
                used_contours.append(cnt)
        except ZeroDivisionError:
            pass

    if cv2.waitKey(1) == 25:
        pass

    return centers


def calc_dist(a, b, pixel_x, pixel_y):
    """
    Calculates the Euclidean distance between two 2D coordinates.
    :param a: Coordinate 1
    :param b: Coordinate 2
    :param pixel_x: pixel size in mm (x-direction)
    :param pixel_y: pixel size in mm (y-direction)
    :return: Distance in pixels and mm, respectively 
    """
    x = a[0] - b[0]
    x_mm = x * pixel_x
    y = a[1] - b[1]
    y_mm = y * pixel_y
    dist = np.sqrt(x ** 2 + y ** 2)
    dist_mm = np.sqrt(x_mm ** 2 + y_mm ** 2)

    return dist, dist_mm


def compute_arc_length(track, pixel_x, pixel_y):
    """
    :param track:
    :param pixel_x: pixel size in mm (x-direction)
    :param pixel_y: pixel size in mm (y-direction)
    :return: 
    """
    length = 0
    for i in range(1, len(track)):
        dist, dist_mm = calc_dist(track[i - 1], track[i], pixel_x, pixel_y)
        length += dist_mm

    return length


def closest_point(center, tracks, pixel_x, pixel_y):
    """
    Find closest detected object in previous frame.
    :param center: Coordinate of current detection.
    :param tracks: Previously calculated tracks to match with the center.
    :param pixel_x: pixel size in mm (x-direction)
    :param pixel_y: pixel size in mm (y-direction)
    :return: closest coordinate in previous frame.
    """
    closest_coord = tracks[0][0][-1]
    closest_i = 0
    min_dist = np.inf
    for i, track in enumerate(tracks):
        dist, dist_mm = calc_dist(center, track[0][-1], pixel_x, pixel_y)
        if dist < min_dist:
            closest_coord = track[0][-1]
            closest_i = i
            min_dist = dist

    return closest_coord, closest_i, min_dist


def find_coordinates(track):
    """
    Find x- and y-coordinates of all tracks. Used for plotting.
    :param track: all tracks throughout the video.
    :return: matrices with x- and y-coordinates.
    """
    x_matrix = []
    y_matrix = []
    times = []
    x = []
    y = []
    time = []
    for element in track[0]:
        x.append(element[0])
        y.append(element[1])
    for element2 in track[1]:
        time.append(element2 / 128.0)
    x_matrix.append(x)
    y_matrix.append(y)
    times.append(time)

    return x_matrix, y_matrix, times


def find_new_track(tracks, centers, pixel_x, pixel_y):
    """
    Create new track if no other track corresponds to current center.
    :param tracks: list of current tracks.
    :param centers: list of centers in current frame.
    :param pixel_x: pixel size in mm (x-direction)
    :param pixel_y: pixel size in mm (y-direction)
    :return: initial coordinate of the new track.
    """

    dist_centers = {}
    for center in centers:
        closest_coord, closest_i, min_dist = closest_point(center, tracks, pixel_x, pixel_y)
        dist_centers[str(center)] = min_dist

    min_dist_coord = min(dist_centers, key=dist_centers.get)
    new_track = ast.literal_eval(min_dist_coord)

    return new_track


def find_tracks(centers, pixel_x, pixel_y):
    """
    Match coordinates of current frame to ongoing tracks, or make new ones.
    :param centers: all centers detected throughout the video.
    :param pixel_x: pixel size in mm (x-direction)
    :param pixel_y: pixel size in mm (y-direction)
    :return: all calculated tracks throughout video.
    """

    """
    Tracks = [[track1], [track2], ..., [trackn]]
    track = [centers, frame_nums, arc_length, distance, det_factor]
    center = [[x1, y1], [x2, y2], ..., [xn, yn]]
    frame_nums = [1, 2, 3, ..., n]
    arc_length = [l]
    distance = [d]
    det_factor = [b, a]
    """
    tracks = []

    start = 0

    while len(centers[start]) == 0:
        start += 1

    for i, center in enumerate(centers[start]):
        tracks.append([[center], [0], [], [], []])

    for j, frame in enumerate(centers, start=start + 1):

        not_updated_tracks = list(tracks)

        for counter, center in enumerate(frame):
            if tracks.__contains__([center]):
                continue
            if len(not_updated_tracks) == 0:
                continue
            closest_coord, closest_i, min_dist = closest_point(center, not_updated_tracks, pixel_x, pixel_y)
            does_exist = False
            if min_dist > 20:  # Change this requirement if wanted
                tracks.append([[center], [int(j)], [], [], []])
                continue
            for track in not_updated_tracks:
                if track[0][-1] == closest_coord:
                    not_updated_tracks.remove(track)
                    tracks.remove(track)
                    new_track = track
                    new_track[0].append(center)
                    new_track[1].append(int(j))
                    tracks.append(new_track)
                    does_exist = True

            if does_exist is False:
                tracks.append([[center], [int(j)], [], [], []])

        if len(not_updated_tracks) > 0:
            for track in not_updated_tracks:
                not_updated_tracks.remove(track)
                tracks.remove(track)
                new_track = track
                new_track[0].append(new_track[0][-1])
                new_track[1].append(int(j))
                tracks.append(new_track)

    print("tracks found")
    # Compute arc length, distance and add to track.
    for i, track in enumerate(tracks):
        # Compute arc length, distance
        arc_length = compute_arc_length(track[0], pixel_x, pixel_y)
        dist, dist_mm = calc_dist(track[0][0], track[0][-1], pixel_x, pixel_y)
        track[2].append(arc_length)
        track[3].append(dist_mm)

        # Compute det factors
        det_factors = calculate_det_factor(track, pixel_x, pixel_y)
        track[4] = [factor for factor in det_factors]

        # Add coordinated in x- and y-direction + speed
        x, y, times_track = find_coordinates(track)
        data_x = [int(f) for f in x[0]]
        data_y = [int(f) for f in y[0]]

        x_coords = []
        y_coords = []

        for element in data_x:
            x_coords.append([element])
        for element in data_y:
            y_coords.append([element])

        speed_list_x, speed_list_y, speed_list_tot, max_speed_frame = compute_speed(x_coords, y_coords,
                                                                                    pixel_x, pixel_y)

        track.append(x_coords)
        track.append(y_coords)
        track.append(speed_list_x)
        track.append(speed_list_y)
        track.append(speed_list_tot)
        track.append([max_speed_frame / 128.0])

        # Add times
        track_times = []
        for t in track[1]:
            track_times.append(t / 128.0)
        track.append(track_times)

    # Remove short tracks
    updated_tracks = []
    for index_track, track_object in enumerate(tracks):
        output = Counter([tuple(i) for i in track_object[0]])
        unique_speed_x = len(set(track_object[7]))
        unique_speed_y = len(set(track_object[8]))
        if float(track_object[2][0]) > 10.0 and len(output) > 128.0 and track_object[5][0][0] > 15.0 \
                and unique_speed_x > 7 and unique_speed_y > 7:
            updated_tracks.append(track_object)

    return updated_tracks


def calculate_det_factor(track, pixel_x, pixel_y):
    """
    Calculated the determination factor of the swimming behavior.
    :param track: the track of one louse.
    :param pixel_x: pixel size in mm (x-direction)
    :param pixel_y: pixel size in mm (y-direction)
    :return: determination factor (distance/arc_length)
    """
    # Values full videos
    time_light_s = 30.0
    time_light_frames = 30.0 * 128

    det_factor_total = track[3][0] / track[2][0]

    track_dark = []
    track_light = []
    if time_light_frames in track[1]:
        index = track[1].index(time_light_frames)
    else:
        index = min(range(len(track[1])), key=lambda ii: abs(track[1][ii] - time_light_frames))

    for i in range(index):
        track_dark.append(track[0][i])

    for j in range(index, len(track[0])):
        track_light.append(track[0][j])

    if len(track_dark) > 5:
        dist_dark, dist_dark_mm = calc_dist(track_dark[0], track_dark[-1], pixel_x, pixel_y)
        arc_length_dark = 0
        for k in range(1, len(track_dark)):
            dist, dist_mm = calc_dist(track_dark[k - 1], track_dark[k], pixel_x, pixel_y)
            arc_length_dark += dist_mm
        det_factor_dark = dist_dark_mm / arc_length_dark
    else:
        det_factor_dark = 0

    if len(track_light) > 5:
        dist_light, dist_light_mm = calc_dist(track_light[0], track_light[-1], pixel_x, pixel_y)
        arc_length_light = 0
        for l in range(1, len(track_light)):
            dist, dist_mm = calc_dist(track_light[l - 1], track_light[l], pixel_x, pixel_y)
            arc_length_light += dist_mm
        det_factor_light = dist_light_mm / arc_length_light
    else:
        det_factor_light = 0

    det_factors = [det_factor_total, det_factor_dark, det_factor_light]

    return det_factors


def compute_speed(x_coords, y_coords, pixel_x, pixel_y):
    """
    Computes the speed of each louse.
    :param x_coords: x-coordinates of track
    :param y_coords: y-coordinates of track
    :param pixel_x: pixel size in mm (x-direction)
    :param pixel_y: pixel size in mm (y-direction)
    :return: total speed, speed in x-direction, speed in y-direction and maximum speed
    """
    x = []
    y = []
    tot = []

    tracks = []

    speed_list_x = [0.0]
    speed_list_y = [0.0]
    speed_list_tot = [0.0]

    dt = 1/(128.0)

    max_speed = 0
    max_speed_frame = np.inf

    for j in range(1, len(x_coords) - 2):
        x_vals = x_coords
        y_vals = y_coords
        speed_x = (((x_vals[j + 1][0]) - (x_vals[j - 1][0]))*pixel_x) / (2*dt)
        speed_y = (((y_vals[j + 1][0]) - (y_vals[j - 1][0]))*pixel_y) / (2*dt)
        speed_tot = math.sqrt(((x_vals[j + 1][0] - x_vals[j - 1][0])*pixel_x) ** 2 +
                      ((y_vals[j + 1][0] - y_vals[j - 1][0])*pixel_y) ** 2) / (2*dt)
        speed_list_x.append(speed_x)
        speed_list_y.append(speed_y)
        speed_list_tot.append(speed_tot)

        if speed_tot > max_speed:
            max_speed = speed_tot
            max_speed_frame = j

    speed_list_x.append(speed_list_x[-1])
    speed_list_x.append(speed_list_x[-1])
    speed_list_y.append(speed_list_y[-1])
    speed_list_y.append(speed_list_y[-1])
    speed_list_tot.append(speed_list_tot[-1])
    speed_list_tot.append(speed_list_tot[-1])

    x.append(speed_list_x)
    y.append(speed_list_y)
    tot.append(speed_list_tot)

    normalized_x = []
    normalized_y = []
    normalized_tot = []

    for i in range(5, len(speed_list_tot)-5):
        normalized_x.append(mean(speed_list_x[i-5:i+5]))
        normalized_y.append(mean(speed_list_y[i-5:i+5]))
        normalized_tot.append(mean(speed_list_tot[i-5:i+5]))

    diff_x = len(speed_list_x)-len(normalized_x)
    diff_y = len(speed_list_y)-len(normalized_y)
    diff_tot = len(speed_list_tot)-len(normalized_tot)

    print("normalized_x", normalized_x)
    if len(normalized_x) > 0:
        normalized_x.extend([normalized_x[-1]] * diff_x)
        normalized_y.extend([normalized_y[-1]] * diff_y)
        normalized_tot.extend([normalized_tot[-1]] * diff_tot)

    return normalized_x, normalized_y, normalized_tot, max_speed_frame


def save_speed_to_csv(x, y, tot, video):
    path = "W:\TrackingVideos/" + video + "/Results"
    # Data that are being plotted
    print("x", x)
    with open(path + "/Speeds_X.csv", "a") as g:
        wr = csv.writer(g)
        wr.writerows(x)
    with open(path + "/Speeds_Y.csv", "a") as f:
        wr = csv.writer(f)
        wr.writerows(y)
    with open(path + "/Speeds_TOT.csv", "a") as h:
        wr = csv.writer(h)
        wr.writerows(tot)


def save_data(tracks, path):
    list_tracks = []
    tracks_coordinates = []

    for i, track in enumerate(tracks):
        dict_track = {}
        dict_track["trackid"] = i
        dict_track["start_time"] = track[11][0]
        dict_track["arc_length"] = track[2][0]
        dict_track["distance"] = track[3][0]
        dict_track["avg_speed"] = mean(track[9])
        dict_track["max_speed"] = max(track[9])
        dict_track["det_factor_total"] = track[4][0]
        dict_track["det_factor_dark"] = track[4][1]
        dict_track["det_factor_light"] = track[4][2]
        dict_track["max_speed_s"] = track[10][0]
        dict_track["speed_tot"] = track[9]
        dict_track["speed_x"] = track[7]
        track_coordinate = track[0]
        list_tracks.append(dict_track)
        tracks_coordinates.append(track_coordinate)

    keys = list_tracks[0].keys()
    with open(path + "/data.csv", "w") as f:
        wr = csv.DictWriter(f, keys)
        wr.writeheader()
        wr.writerows(list_tracks)

    with open(path + "/tracks.csv", "w") as g:
        wr2 = csv.writer(g)
        wr2.writerows(tracks_coordinates)


def plotting(tracks, time_light, video, pixel_x, pixel_y):
    """
    Plots the movement of objects over time.
    :param tracks: all tracks throughout the video.
    :param time_light: time when light is turned on for the first time.
    :param video: video file being analyzed
    :param pixel_x: size of pixel in x-direction
    :param pixel_y: size of pixel in y-direction
    :return: -
    """

    start = time.time()

    # Define colors
    color_palette = sns.color_palette("deep", 12)
    track_colors = []
    for color in color_palette:
        rgb_color = ()
        for value in color:
            rgb_color += (value,)
        track_colors.append(rgb_color)

    path = "W:\TrackingVideos/" + video + "/Results"

    num_frames = get_information(video)

    list_num_frames = np.arange(num_frames)

    times = []

    for element in list_num_frames:
        times.append(element / 128.0)

    for i, track in enumerate(tracks):
        index = i % 12
        color = track_colors[index]

        x_text = track[5][0][0] * pixel_x + 0.3
        y_text = track[6][0][0] * pixel_y + 0.3
        time_text = track[11][0] + 3

        plt.figure(1, figsize=(20, 7))
        plt.plot(track[11], [[b * pixel_x for b in a] for a in track[5]], color=color)
        plt.axvline(x=time_light, linestyle='--', color='k', linewidth=0.1)
        plt.title("X position vs time, {}".format(video))
        plt.xlabel("Time [s]")
        plt.ylabel("X [mm]")
        plt.savefig(path + "/X_vs_time_" + video + ".png")

        plt.figure(2, figsize=(20, 7))
        plt.plot(track[11], [[x * pixel_y for x in d] for d in track[6]], color=color)
        plt.axvline(x=time_light, linestyle='--', color='k', linewidth=0.1)
        plt.title("Y position vs time, {}".format(video))
        plt.xlabel("Time [s]")
        plt.ylabel("Y [mm]")
        plt.savefig(path + "/Y_vs_time_" + video + ".png")

        plt.figure(3, figsize=(20, 16))
        plt.plot([n * pixel_x for n in track[5][0]], [m * pixel_y for m in track[6][0]], color=color, marker="*",
                 markersize=7)
        plt.axis([0, 210, 0, 136])
        plt.subplots_adjust(bottom=0.4)
        plt.text(x_text, y_text, "{}".format(i, markersize=0.1, fontsize=4))
        label = "{}".format(i) + ": Start: {0:.2f}".format(track[11][0]) + \
                "s, Arc length: {0:.2f}".format(track[2][0]) + "mm, Avg speed: {0:.2f}".format(mean(track[9])) + \
                "mm/s, Max speed: {0:.2f}".format(max(track[9])) + "mm/s"
        plt.plot([[l * pixel_x for l in k] for k in track[5]], [[o * pixel_y for o in u] for u in track[6]], color=color,
                 marker=".", markersize=0.05,
                 label=label)
        plt.title("X vs Y, {}".format(video))
        plt.xlabel("X [mm]")
        plt.ylabel("Y [mm]")
        plt.legend(loc=9, bbox_to_anchor=(0.5, -0.1), prop={'size': 14})
        plt.savefig(path + "/X_vs_Y_bottom_" + video + ".png")

        plt.figure(4, figsize=(23, 10))
        plt.plot([s * pixel_x for s in track[5][0]], [t * pixel_y for t in track[6][0]], color=color, marker="*",
                 markersize=7)
        plt.subplots_adjust(right=0.7)
        plt.text(x_text, y_text, "{}".format(i, markersize=0.1, fontsize=4))
        label = "{}".format(i) + ": Start: {0:.2f}".format(track[11][0]) + \
                "s, Arc length: {0:.2f}".format(track[2][0]) + "mm, Avg speed: {0:.2f}".format(mean(track[9])) + \
                "mm/s, Max speed: {0:.2f}".format(max(track[9])) + "mm/s"
        plt.plot([[y * pixel_x for y in q] for q in track[5]], [[h * pixel_y for h in g] for g in track[6]], color=color,
                 marker=".", markersize=0.05,
                 label=label)
        plt.title("X vs Y, {}".format(video))
        plt.xlabel("X [mm]")
        plt.ylabel("Y [mm]")
        plt.legend(loc='upper left', bbox_to_anchor=(1.01, 1), prop={'size': 9})
        plt.savefig(path + "/X_vs_Y_right_" + video + ".png")

        plt.figure(5, figsize=(20, 10))
        plt.plot([w * pixel_x for w in track[5][0]], [v * pixel_y for v in track[6][0]], color=color, marker="*",
                 markersize=7)
        plt.text(x_text, y_text, "{}".format(i, markersize=0.1, fontsize=4))
        plt.plot([[c * pixel_x for c in g] for g in track[5]], [[e * pixel_y for e in o] for o in track[6]], color=color,
                 marker=".", markersize=0.05)
        plt.title("X vs Y, {}".format(video))
        plt.xlabel("X [mm]")
        plt.ylabel("Y [mm]")
        plt.savefig(path + "/X_vs_Y_without_" + video + ".png")

        plt.figure(7 * (i + 1), figsize=(15, 5))
        plt.plot(track[11], track[7], color=color)
        plt.axvline(x=time_light, linestyle='--', color='k', linewidth=0.1)
        plt.title("Speed in X direction, track " + str(i))
        plt.ylabel("u [mm/s]")
        plt.xlabel("Time [s]")
        plt.savefig(path + "/X_speed_" + str(i) + ".png")

        plt.figure(90 * (i + 1), figsize=(15, 5))
        plt.plot(track[11], track[8], color=color)
        plt.axvline(x=time_light, linestyle='--', color='k', linewidth=0.1)
        plt.title("Speed in Y direction, track " + str(i))
        plt.ylabel("v [mm/s]")
        plt.xlabel("Time [s]")
        plt.savefig(path + "/Y_speed_" + str(i) + ".png")

        plt.figure(1100 * (i + 1), figsize=(15, 5))
        plt.plot(track[11], track[9], color=color)
        plt.axvline(x=time_light, linestyle='-', color='k', linewidth=0.1)
        plt.title("Total speed, track " + str(i))
        plt.ylabel("V [mm/s]")
        plt.xlabel("Time [s]")
        plt.savefig(path + "/tot_speed" + str(i) + ".png")

    end = time.time()
    print("Time plotting: ", end - start)

    return 0


def main():
    """Pipeline:
    - Make a list of sorted frames
    - Calculate median of all frames, distinct medians for bright and dark frames
    - Loop through all frames, one frame at the time
        - Choose median depending on value of brightness in frame
        - Detect and retrieve coordinates of the centers of objects in the frame
    - Use center coordinates in all frames to retrieve individual tracks for each object
    """

    start_main = time.time()
    
    # Update length of one pixel in x- and y-direction, unit: mm. 
    pixel_x = 0.090909
    pixel_y = 0.080769

    list_videos = ["T_Blue_OD1_P0_R1", "T_Blue_OD1_P0_R2", "T_Blue_OD1_P0_R3",
                   "T_White_OD1_P3_R1", "T_White_OD1_P3_R2", "T_White_OD1_P3_R2",
                   "T_Green_OD1_P2_R2", "T_Green_OD1_P2_R1", "T_Green_OD1_P2_R3"]

    for video in list_videos:

        print("Current video: ", video)

        path_res = "W:\TrackingVideos/" + video + "/Results"

        if os.path.exists(path_res):
            import shutil
            shutil.rmtree(path_res)

        os.makedirs(path_res)

        path = "W:\TrackingVideos/" + video

        frames = os.listdir(path)

        if frames.__contains__(".DS_Store"):
            frames.remove(".DS_Store")

        if frames.__contains__("Results"):
            frames.remove("Results")

        frames.sort()

        all_centers = []

        a = video.split("_")
        pulse_type = a[3]

        if pulse_type == "P0":
            median_dark, median_light = calculate_median_p0(video)
        else:
            median, median_bright, median_dark, min_mean, max_mean = calculate_median(video)

        # For hele forsoket
        time_light = 128 * 30

        for i, frame in enumerate(frames):
            if i < 10:
                continue
            path_frame = os.path.join(path, frame)

            with open(path_frame, "rb") as f:
                raw = f.read()
                im = np.frombuffer(raw, dtype=np.uint8).reshape(1560, 2200).astype(np.int16)
                im = im.astype(np.uint8)
                im = im[200:1400, 315:2095]

                if pulse_type == "P0":
                    median = choose_median_p0(i, median_dark, median_light)
                else:
                    median = choose_median(im, median_bright, median_dark, min_mean, max_mean, video)

                centers = detect(im, median)

                all_centers.append(centers)

                # Check for key strokes
                if cv2.waitKey(1) == 25:
                    break

        print("centers are found")
        if time_light == np.inf:
            time_light = 0.0
        else:
            time_light = time_light / 128.0

        with open(path_res + "/centers.csv", "w") as i:
            wr = csv.writer(i)
            wr.writerows(all_centers)

        tracks = find_tracks(all_centers, pixel_x, pixel_y)
        print("tracks are found")

        plotting(tracks, time_light, video, pixel_x, pixel_y)
        print("plotting finished")

        save_data(tracks, path_res)
        print("data saved")

    end_main = time.time()

    print("Execution time: ", end_main - start_main)


if __name__ == "__main__":
    # execute main
    main()